// [v0.1.060-20250329]

// === module init block ===

const {
  isArray, isPlainObject,
} = require('./lib-hfunc');

// === module extra block (helper functions) ===

// === module main block ===

/***
 * (* constant definitions *)
 */

/**
 * a ref to `Number.isInteger()` method
 * @param {any} value - some value to verify
 * @returns {boolean}
 */
const isInteger = Number.isInteger;

/**
 * @inner
 * @description A base map for ref from a string value to a boolean value.
 */
const mpStrToBoolRefBase = new Map([
  [ 'false', false ],
  [ 'FALSE', false ],
  [ 'true', true ],
  [ 'TRUE', true ],
]);

/**
 * @inner
 * @description A extended map for ref from a string value to a boolean value.
 */
const mpStrToBoolRefEx = new Map([
  [ 'false', false ],
  [ 'FALSE', false ],
  [ 'true', true ],
  [ 'TRUE', true ],
  [ 'null', false ],
  [ 'NULL', false ],
  [ 'off', false ],
  [ 'OFF', false ],
  [ 'on', true ],
  [ 'ON', true ],
  [ 'no', false ],
  [ 'No', false ],
  [ 'NO', false ],
  [ 'yes', true ],
  [ 'Yes', true ],
  [ 'YES', true ],
]);

/***
 * (* function definitions *)
 */

/**
 * @function valueToIndex
 * @param {any} value - some value
 * @returns {number}
 * @description Converts a given value to an index value.
 *              If failed a `-1` is returned.
 */
function valueToIndex(value) {
  let index = value;
  let result = -1;
  switch (typeof index) {
    case 'string': {
      if ((index = index.trim()) === '') break;
      index = Number(index);
    }
    case 'number': {
      if (Number.isNaN(index)) break;
      if (index < 0 || !isInteger(index)) break;
      result = index;
      break;
    }
    default: {}
  };
  return result;
};

/**
 * @function isValidIndex
 * @param {any} value - some value to verify
 * @returns {boolean}
 * @description Checks whether a given value can be converted
 *              to an index value.
 * @see valueToIndex
 */
function isValidIndex(value) {
  return valueToIndex(value) !== -1;
};

/**
 * @function isEmptyString
 * @param {any} value - some value to verify
 * @returns {boolean}
 * @description Checks whether a given value is an empty string.
 */
function isEmptyString(value) {
  if (typeof value !== 'string') return false;
  return value === '' || value.trim() === '';
};

/**
 * @since 0.1.4
 * @function isEmptyOrNotString
 * @param {any} value - some value to verify
 * @returns {boolean}
 * @description Checks whether a given value is an empty string
 *              or not a string at all.
 */
function isEmptyOrNotString(value) {
  if (typeof value !== 'string') return true;
  return value === '' || value.trim() === '';
};

/**
 * @function isNotEmptyString
 * @param {any} value - some value to verify
 * @returns {boolean}
 * @description Checks whether a given value is not an empty string.
 */
function isNotEmptyString(value) {
  if (typeof value !== 'string') return false;
  return value !== '' && value.trim() !== '';
};

/**
 * @function isNotEmptyStringEx
 * @param {...any} args - some values to verify
 * @returns {boolean}
 * @experimental
 * @description Checks whether all of given values are not an empty string.
 */
function isNotEmptyStringEx(...args) {
  let isSUCCEED = false;
  for (let value of args) {
    if (isArray(value)) {
      isSUCCEED = isNotEmptyStringEx(...value);
    } else {
      isSUCCEED = isNotEmptyString(value);
    };
    if (!isSUCCEED) break;
  };
  return isSUCCEED;
};

/**
 * @typedef {Object} OPT_valtoid
 * @property {boolean} [onlyNES=true] - a flag to accept only a non-empty strings
 * @property {boolean} [ignoreNumbers=false] - a flag to ignore numbers
 * @property {boolean} [ignoreCase=false] - a flag to ignore character case
 * @description an option set
 */

/**
 * @since 0.1.3
 * @function valueToIDString
 * @param {any} value - some value
 * @param {(boolean|OPT_valtoid)} [opt={}] - options
 * @returns {string}
 * @description Converts a given value to an ID.
 *              If failed `null` is returned.
 * @todo [since `v0.2.3`] deprecate use of `opt` as `boolean`
 */
function valueToIDString(value, opt = {}) {
  let result = null;
  /** @type {OPT_valtoid} */
  let {
    onlyNES,
    ignoreNumbers,
    ignoreCase,
  } = isPlainObject(opt) ? opt : { onlyNES: opt };
  if (typeof onlyNES !== 'boolean') onlyNES = true;
  if (typeof ignoreNumbers !== 'boolean') ignoreNumbers = false;
  if (typeof ignoreCase !== 'boolean') ignoreCase = false;
  switch (typeof value) {
    case 'string': {
      let id = value.trim();
      if (id === '') {
        if (onlyNES) break;
      } else {
        const num = Number(id);
        if (Number.isNaN(num)) {
          if (ignoreCase) id = id.toLowerCase();
        } else {
          if (ignoreNumbers) break;
          if (num < 0 || !isInteger(num)) break;
        };
      };
      result = id;
      break;
    }
    case 'number': {
      if (!ignoreNumbers) {
        let id = valueToIndex(value);
        if (id !== -1) result = id.toString();
      };
      break;
    }
    default: {}
  };
  return result;
};

/**
 * @function isValidId
 * @param {any} value - some value to verify
 * @param {(boolean|OPT_valtoid)} [opt={}] - options
 * @returns {boolean}
 * @experimental
 * @description Checks whether a given value can be converted to an ID.
 * @see valueToIDString
 */
function isValidId(...args) {
  return valueToIDString(...args) !== null;
};

/**
 * @function isNullOrUndef
 * @param {any} value - some value to verify
 * @returns {boolean}
 * @description Checks whether a given value is an `undefined` or a `null`.
 */
function isNullOrUndef(value = null) {
  return value === null;
};

/**
 * @since 0.2.1
 * @function tryToBool
 * @param {any} value - some value
 * @returns {?boolean}
 * @description Tries to read a given value as a boolean type.
 *              If failed `null` is returned.
 */
function tryToBool(value) {
  switch (typeof value) {
    case 'boolean': {
      return value;
    }
    case 'string': {
      const result = mpStrToBoolRefBase.get(value.trim());
      if (result !== undefined) return result;
      break;
    }
    default: {}
  };
  return null;
};

/**
 * @function readAsBool
 * @param {any} value - some value
 * @param {boolean} [defValue=false] - some default value to use
 * @returns {boolean}
 * @description Reads a given value and translates it to a boolean type.
 */
function readAsBool(value, defValue) {
  let result = tryToBool(value);
  if (result === null) {
    result = typeof defValue === 'boolean' ? defValue : false;
  };
  return result;
};

/**
 * @since 0.2.1
 * @function tryToNumber
 * @param {any} value - some value
 * @returns {?number}
 * @description Tries to read a given value as a number type.
 *              If failed `null` is returned.
 */
function tryToNumber(value) {
  let result = value;
  switch (typeof value) {
    case 'string': {
      if ((result = result.trim()) === '') break;
      result = Number(result);
    }
    case 'number': {
      if (!Number.isNaN(result)) return result;
      break;
    }
    default: {}
  };
  return null;
};

/**
 * @function readAsNumber
 * @param {any} value - some value
 * @param {number} [defValue=0] - some default value to use
 * @returns {number}
 * @description Reads a given value and translates it to a number type.
 */
function readAsNumber(value, defValue) {
  let result = tryToNumber(value);
  if (result === null) {
    result = (
      typeof defValue === 'number' && !Number.isNaN(defValue)
      ? defValue
      : 0
    );
  };
  return result;
};

/**
 * @typedef {Object} OPT_valtostr
 * @property {boolean} [useTrim=false] - a flag to use `trim`-option
 * @property {string} [defValue=''] - a default value to use
 * @property {boolean} [numberToString=false] - a flag to transform numbers to string
 * @property {boolean} [boolToString=false] - a flag to transform booleans to string
 * @property {string} [letterCase] - a mode to transform a letter case
 * @property {boolean} [boolToUpperCase=false] - a flag to transform a letter case for a booleans
 */

/**
 * @function readAsString
 * @param {any} value - some value
 * @param {(boolean|string|OPT_valtostr)} [opt] - options
 * @returns {string}
 * @description Reads a given value and translates it to a string type.
 * @todo [since `v0.2.2`] deprecate use of `opt` as `boolean`
 * @todo [since `v0.2.2`] deprecate use of `opt` as `string`
 */
function readAsString(value, opt) {
  let _options = opt;
  switch (typeof _options) {
    case 'boolean': {
      _options = { useTrim: _options };
      break;
    }
    case 'string': {
      _options = { defValue: _options };
      break;
    }
    default: {
      if (!isPlainObject(_options)) _options = {};
    }
  };
  let useUpperCase = false;
  let useLowerCase = false;
  let {
    useTrim,
    letterCase,
    numberToString,
    boolToString,
    boolToUpperCase,
    defValue,
  } = _options;
  if (typeof useTrim !== 'boolean') useTrim = false;
  if (typeof numberToString !== 'boolean') numberToString = false;
  if (typeof boolToString !== 'boolean') boolToString = false;
  if (typeof boolToUpperCase !== 'boolean') boolToUpperCase = false;
  switch (letterCase) {
    case 'upper': {
      useUpperCase = true;
      boolToUpperCase = true;
      break;
    }
    case 'lower': {
      useLowerCase = true;
      break;
    }
    default: {}
  };
  if (typeof defValue !== 'string') defValue = '';
  let result = '';
  if (numberToString && typeof value === 'number') {
    if (!Number.isNaN(value)) return value.toString();
  } else if (boolToString && typeof value === 'boolean') {
    result = value ? 'true' : 'false';
    return boolToUpperCase ? result.toUpperCase() : result;
  };
  result = typeof value === 'string' ? value : defValue;
  if (useTrim) result = result.trim();
  if (useUpperCase) result = result.toUpperCase();
  if (useLowerCase) result = result.toLowerCase();
  return result;
};

/**
 * @function readAsBoolEx
 * @param {any} value - some value
 * @param {boolean} [defValue=false] - some default value to use
 * @returns {boolean}
 * @description Reads a given value and translates it to a boolean type.
 */
function readAsBoolEx(value, defValue) {
  let _value = value;
  if (_value === null) return false;
  switch (typeof _value) {
    case 'boolean': {
      return _value;
    }
    case 'string': {
      _value = _value.trim();
      const result = mpStrToBoolRefEx.get(_value);
      if (result !== undefined) return result;
      _value = Number(_value);
    }
    case 'number': {
      if (!Number.isNaN(_value)) return _value !== 0;
    }
    default: {}
  };
  return typeof defValue === 'boolean' ? defValue : false;
};

/**
 * @function readAsNumberEx
 * @param {any} value - some value
 * @param {number} [defValue=0] - some default value to use
 * @returns {number}
 * @description Reads a given value and translates it to a number type.
 */
function readAsNumberEx(value, defValue) {
  let _value = value;
  if (_value === null) return 0;
  switch (typeof _value) {
    case 'boolean': {
      return _value ? 1 : 0;
    }
    case 'string': {
      _value = _value.trim();
      const result = mpStrToBoolRefEx.get(_value);
      if (result !== undefined) return result ? 1 : 0;
      _value = Number(_value);
    }
    case 'number': {
      if (!Number.isNaN(_value)) return _value;
    }
    default: {}
  };
  return (
    typeof defValue === 'number' && !Number.isNaN(defValue)
    ? defValue
    : 0
  );
};

/**
 * @function tryNotNullish
 * @param {...any} args - some values
 * @returns {?any}
 * @description Reads a list of a given arguments and returns
 *              a first one that value is not a `null` or `undefined`.
 */
function tryNotNullish(...args) {
  let result = null;
  let value = undefined;
  for (value of args) {
    if (!isNullOrUndef(value)) {
      result = value;
      break;
    };
  };
  return result;
};

/**
 * @function inRangeNum
 * @param {number} minR - minimum
 * @param {number} maxR - maximum
 * @param {number} value - some value to verify
 * @returns {boolean}
 * @description Checks whether a given value is into a defined range.
 */
function inRangeNum(minR, maxR, value) {
  return value >= minR && value <= maxR;
};

/***
 * (* class definitions *)
 */

// === module exports block ===

module.exports.isInteger = isInteger;
module.exports.isValidIndex = isValidIndex;
module.exports.isEmptyString = isEmptyString;
module.exports.isEmptyOrNotString = isEmptyOrNotString;
module.exports.isNotEmptyString = isNotEmptyString;
module.exports.isNullOrUndef = isNullOrUndef;
module.exports.inRangeNum = inRangeNum;

module.exports.valueToIndex = valueToIndex;
module.exports.valueToIDString = valueToIDString;
module.exports.readAsBool = readAsBool;
module.exports.readAsBoolEx = readAsBoolEx;
module.exports.readAsNumber = readAsNumber;
module.exports.readAsNumberEx = readAsNumberEx;
module.exports.readAsString = readAsString;
module.exports.tryNotNullish = tryNotNullish;
module.exports.tryToBool = tryToBool;
module.exports.tryToNumber = tryToNumber;

// * experimental *
module.exports.isNotEmptyStringEx = isNotEmptyStringEx;
module.exports.isValidId = isValidId;
