// [v0.1.084-20250329]

// === module init block ===

const {
  readAsBool, readAsNumber, readAsString,
  isNullOrUndef, isNotEmptyString,
} = require('./baseFunc.js');

const {
  isArray, isObject, isPlainObject,
} = require('./lib-hfunc');


// === module extra block (helper functions) ===

// === module main block ===

/***
 * (* constant definitions *)
 */

/**
 * a ref to `Array.from()` method
 * @returns {any[]}
 */
const arrayFrom = Array.from;

/***
 * (* function definitions *)
 */

/**
 * @function isDateObject
 * @param {any} value - some value to verify
 * @returns {boolean}
 * @description Checks whether a given value is a `DateTime`-object.
 */
function isDateObject(value) {
  return Object.prototype.toString.call(value) === '[object Date]';
};

/**
 * @function valueToArray
 * @param {any} value - some value
 * @returns {any[]}
 * @description Transforms a given value to an array.
 */
function valueToArray(value) {
  if (isNullOrUndef(value)) return [];
  if (isArray(value)) return value;
  if (
    value instanceof Set || value instanceof Map
  ) {
    return arrayFrom(value);
  };
  return [ value ];
}

/**
 * @callback mapArrFn
 * @param {any} item - some element
 * @param {number} index - an element index
 * @returns {any}
 * @description a user defined procedure to pick up an array elements
 */

/**
 * @since 0.2.3
 * @typedef {Object} OPT_valtoarrex
 * @property {mapArrFn} [mapFn] - a user defined procedure
 * @property {boolean} [treatMapsAsObject=false] - a flag to keep 'Map'-object
 * @description An options set
 */

/**
 * @since 0.1.1
 * @function valueToArrayEx
 * @param {any} value - some value
 * @param {(mapArrFn|OPT_valtoarrex)} [opt] - some callback function or an options
 * @returns {any[]}
 * @experimental
 * @description Transforms a given value to an array.
 */
function valueToArrayEx(value = [], opt) {
  /** @type {OPT_valtoarrex} */
  let {
    mapFn,
    treatMapsAsObject,
  } = isPlainObject(opt) ? opt : { mapFn: opt };
  if (typeof mapFn !== 'function') mapFn = undefined;
  if (typeof treatMapsAsObject !== 'boolean') treatMapsAsObject = false;
  if (
    value instanceof Set
    || (!treatMapsAsObject && value instanceof Map)
  ) {
    return arrayFrom(value, mapFn);
  };
  return arrayFrom(isArray(value) ? value : [ value ], mapFn);
}

/**
 * @since 0.2.1
 * @function valueToEntry
 * @param {any} value - some value
 * @returns {?array}
 * @description Transforms a given value to an entry (aka `Key-Value`-pair).
 *              If failed `null` is returned.
 */
function valueToEntry(value = null) {
  if (value === null) return null;
  let [ key, _value ] = isArray(value) ? value : [ '', value ];
  if (isNullOrUndef(key)) return null;
  if (typeof key === 'string') key = key.trim();
  return [ key, _value ];
};

/**
 * @since 0.1.4
 * @function valueToEntries
 * @param {any} value - some value
 * @param {object} [opt]
 * @returns {any[]}
 * @description Transforms a given value to an array of entries.
 */
function valueToEntries(value, opt) {
  let _options = isPlainObject(opt) ? opt : {};
  let result = [];
  valueToArray(value).forEach((item) => {
    const entry = valueToEntry(item);
    if (entry !== null) {
      let [ key, value ] = entry;
      if (typeof key === 'string') {
        value = readAsString(value, _options);
        if (value !== '') result.push([ key, value ]);
      };
    };
  });
  return result;
};

/**
 * @typedef {Object} arrManipSetDesc
 * @property {function} [checkItemFn] - a func to check element
 * @property {function} [readItemFn] - a func to read element
 * @description a set of an array manipulators
 */

/**
 * @typedef {Object} OPT_readliste
 * @property {boolean} [keepEmpty=false] - a flag that defines whether to keep empties
 * @property {boolean} [unfoldArray=false] - a flag that defines whether to unfold an arrays
 * @property {arrManipSetDesc} [userFn] - a set of a user defined manips
 * @property {any} [userOptions] - <*reserved*>
 * @description an options set
 */

/**
 * @function readAsListE
 * @param {(boolean|OPT_readliste)} opt - options
 * @param {...any} args
 * @returns {any[]}
 * @description Reads a list of an elements.
 * @todo [since `v0.2.3`] deprecate use of `opt` as `boolean`
 */
function readAsListE(opt, ...args) {
  const __retItem = function(item) {
    return isPlainObject(item) ? item : null;
  };
  const __chkItem = function(item, opt) {
    return readAsBool(opt, false) ? true : (item ? true : false);
  };
  /** @type {OPT_readliste} */
  let {
    keepEmpty,
    unfoldArray,
    userFn,
    userOptions,
  } = isPlainObject(opt) ? opt : { keepEmpty: opt };
  if (typeof keepEmpty !== 'boolean') keepEmpty = false;
  if (typeof unfoldArray !== 'boolean') unfoldArray = false;
  if (!isPlainObject(userFn)) userFn = {};
  let { checkItemFn, readItemFn } = userFn;
  if (typeof readItemFn !== 'function') readItemFn = __retItem;
  if (typeof checkItemFn !== 'function') checkItemFn = __chkItem;
  /**
   * reads an array elements
   * @param {any[]} arr - source
   * @param {any[]} result - receiver
   * @param {function} readFn - a func to read an element
   * @param {function} checkFn - a func to check an element
   * @param {boolean} keepEmpty - a flag to keep empties
   * @param {boolean} unfoldArray - a flag to unfold
   * @returns {any[]}
   */
  const __readElements = (
    arr, result, readFn, checkFn, keepEmpty, unfoldArray
  ) => {
    for (let elem of arr) {
      if (unfoldArray && isArray(elem)) {
        __readElements(elem, result, readFn, checkFn, keepEmpty, false);
      } else if (unfoldArray && (elem instanceof Set)) {
        __readElements([...elem], result, readFn, checkFn, keepEmpty, false);
      } else {
        const value = readFn(elem);
        if (value !== undefined && checkFn(value, keepEmpty)) {
          result.push(value);
        };
      };
    };
    return result;
  };
  return __readElements(
    [ ...args ], [], readItemFn, checkItemFn, keepEmpty, unfoldArray,
  );
};

/**
 * @typedef {Object} OPT_readlists
 * @property {boolean} [keepEmpty=false] - a flag that defines whether to keep empties
 * @property {boolean} [unfoldArray=false] - a flag that defines whether to unfold an arrays
 * @property {number} [unfoldLevel=0] - defines a level of  an arrays nesting to unfold
 * @property {any} [userOptions] - some user options
 * @description an options set
 */

/**
 * @function readAsListS
 * @param {(boolean|string|OPT_readlists)} opt - options
 * @param {...any} arg - some element to pick up
 * @returns {string[]}
 * @description Reads a list of an elements.
 * @todo [since `v0.2.2`] deprecate use of `opt` as `boolean`
 * @todo [since `v0.2.2`] deprecate use of `opt` as `string`
 */
function readAsListS(opt, ...args) {
  /** @type {OPT_readlists} */
  let {
    keepEmpty,
    unfoldArray,
    unfoldLevel,
    userOptions,
  } = isPlainObject(opt) ? opt : { userOptions: opt };
  if (typeof keepEmpty !== 'boolean') keepEmpty = false;
  if (typeof unfoldArray !== 'boolean') unfoldArray = false;
  if (unfoldArray) {
    unfoldLevel = readAsNumber(unfoldLevel, 1);
    if (unfoldLevel < -1) unfoldLevel = 0;
  } else {
    unfoldLevel = 0;
  };
  /**
   * reads an array elements
   * @param {any[]} arr - source
   * @param {string[]} result - receiver
   * @param {any} opt - options
   * @param {boolean} keepEmpty - a flag to keep empties
   * @param {number} unfoldLevel - a level of nesting
   * @returns {string[]}
   */
  const __readElements = (
    arr, result, opt, keepEmpty, unfoldLevel,
  ) => {
    const unfoldArray = unfoldLevel > 0 || unfoldLevel === -1;
    for (let elem of arr) {
      if (unfoldArray && isArray(elem)) {
        if (unfoldLevel > 0) unfoldLevel--;
        __readElements(elem, result, opt, keepEmpty, unfoldLevel);
      } else {
        const value = readAsString(elem, opt);
        if (keepEmpty || isNotEmptyString(value)) result.push(value);
      };
    };
    return result;
  };
  return __readElements([ ...args ], [], userOptions, keepEmpty, unfoldLevel);
};

/**
 * @function isPropertyExists
 * @param {object} obj - some instance
 * @param {string} name - a property name
 * @returns {boolean}
 * @description Checks whether an object has a target property.
 * @todo consider if an array is acceptable as "obj"
 */
function isPropertyExists(obj, name) {
  let result = false;
  let key = '';
  if (
    isObject(obj)
    && typeof name === 'string'
    && ((key = name.trim()) !== '')
  ) {
    switch (typeof obj[key]) {
      case 'undefined':
      case 'function': {
        result = false;
        break;
      }
      default: {
        result = true;
      }
    };
  };
  return result;
};

/**
 * @function isMethodExists
 * @param {object} obj - some instance
 * @param {string} name - a method name
 * @returns {boolean}
 * @description Checks whether an object has a target method.
 * @todo consider if an array is acceptable as "obj"
 */
function isMethodExists(obj, name) {
  let result = false;
  let key = '';
  if (
    isObject(obj)
    && typeof name === 'string'
    && ((key = name.trim()) !== '')
  ) {
    result = typeof obj[key] === 'function';
  };
  return result;
};

/**
 * @function doMethodIfExists
 * @param {object} obj - some instance
 * @param {string} name - a method name
 * @param {...any} args - a method arguments
 * @returns {any}
 * @experimental
 * @description Performs a target method of an object if it's exists.
 */
function doMethodIfExists(obj, name, ...args) {
  if (isMethodExists(obj, name)) return obj[name.trim()](...args);
};

/***
 * (* class definitions *)
 */

// === module exports block ===

module.exports.isArray = isArray;
module.exports.isObject = isObject;
module.exports.isPlainObject = isPlainObject;
module.exports.isDateObject = isDateObject;
module.exports.isPropertyExists = isPropertyExists;
module.exports.isMethodExists = isMethodExists;

module.exports.arrayFrom = arrayFrom;
module.exports.valueToArray = valueToArray;
module.exports.valueToEntry = valueToEntry;
module.exports.valueToEntries = valueToEntries;
module.exports.readAsListE = readAsListE;
module.exports.readAsListS = readAsListS;

// * experimental *
module.exports.doMethodIfExists = doMethodIfExists;
module.exports.valueToArrayEx = valueToArrayEx;
